package ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0

import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.HAS_ROWS
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.U8
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.selectOneOrError
import ru.landgrafhomyak.db.serdha0.api.misc.RowExpression
import ru.landgrafhomyak.db.serdha0.api.misc._Selectable
import ru.landgrafhomyak.db.serdha0.api.queries.SelectQuery
import ru.landgrafhomyak.db.serdha0.api.queries._Query
import ru.landgrafhomyak.db.serdha0.api.runtime.OutputRow
import ru.landgrafhomyak.db.serdha0.api.runtime.Transaction
import ru.landgrafhomyak.db.serdha0.api.table.ForeignRowReference
import ru.landgrafhomyak.db.serdha0.api.table.Table
import ru.landgrafhomyak.db.serdha0.api.table.TableConstructor

@Suppress("ClassName", "PropertyName", "MemberVisibilityCanBePrivate")
class T_User2Queue private constructor(
	context: TableConstructor.Scope<T_User2Queue>,
	usersTable: Table<T_Users, *>, queueSetsTable: Table<T_QueueSets, *>, queuesTable: Table<T_Queues, *>
) {
	val C_RowId = context.rowIdColumn
	val C_User = context.foreignRowReference("user", usersTable, ForeignRowReference.OnDelete.CASCADE)
	val C_QueueSet = context.foreignRowReference("queue_set", queueSetsTable, ForeignRowReference.OnDelete.CASCADE)
	val C_Queue = context.foreignRowReference("queue", context.types.nullableOf(context.types.ROW_ID(queuesTable)), queuesTable, ForeignRowReference.OnDelete.CASCADE)
	val C_Ordinal = context.column("ordinal", context.types.U8)

	val U_UserInQueueSet = context.uniqueIndex("user_in_queue_set", true, this.C_User, this.C_QueueSet)
	val U_UserInQueue = context.uniqueIndex("user_in_queue", true, this.C_QueueSet, this.C_Queue, this.C_Ordinal)

	val H_QueueRef = context.checkConstraint("queue_ref", CheckQueueRefs(queueSetsTable))

	private class CheckQueueRefs(
		private val queueSetsTable: Table<T_QueueSets, *>
	) : RowExpression.Constructor<T_User2Queue, CheckQueueRefs._RowExports, Boolean> {
		override val rowExports = _RowExports

		override fun createRowExpression(internalQueries: _Query.Constructor._SelectsScope, runtimeQueries: _Query.Constructor._SelectsScope): RowExpression.Action<_RowExports, Boolean> {
			val selectQueuesCount = runtimeQueries.selectFolding(_SelectQueuesCount(this.queueSetsTable))
			return _Action(selectQueuesCount)
		}

		private class _RowExports(context: RowExpression.RowExportsConstructor.Scope<_RowExports, T_User2Queue>) {
			val C_QueueSet = context.selectFromTargetTable(context.tt.uExt.C_QueueSet)
			val C_Queue = context.selectFromTargetTable(context.tt.uExt.C_Queue)

			companion object Constructor : RowExpression.RowExportsConstructor<_RowExports, T_User2Queue> {
				override fun createRowExports(context: RowExpression.RowExportsConstructor.Scope<_RowExports, T_User2Queue>) =
					_RowExports(context)
			}
		}

		private class _SelectQueuesCount private constructor(context: SelectQuery.Folding.Constructor.Scope<_SelectQueuesCount, T_QueueSets>) {
			val P_QueueSetId = context.queryParam("queue_set_id", context.types.ROW_ID(context.src as Table<T_QueueSets, *>))
			val C_HasQueries = context.returnColumn("has_queues", context.HAS_ROWS())


			private class _Constructor(queueSetsTable: Table<T_QueueSets, *>) : SelectQuery.Folding.Constructor<_SelectQueuesCount, T_QueueSets> {
				override val source: _Selectable<T_QueueSets> = queueSetsTable

				override fun createSelectWithFold(context: SelectQuery.Folding.Constructor.Scope<_SelectQueuesCount, T_QueueSets>) =
					_SelectQueuesCount(context)
			}

			companion object Constructor {
				operator fun invoke(queueSetsTable: Table<T_QueueSets, *>) =
					_Constructor(queueSetsTable)
			}
		}

		private class _Action(private val selectQueuesCount: SelectQuery.Folding<_SelectQueuesCount>) : RowExpression.Action<_RowExports, Boolean> {
			override suspend fun calculateRow(rowUE: _RowExports, row: OutputRow<_RowExports>, transaction: Transaction): Boolean {
				val hasRows: Boolean = transaction.selectOneOrError(
					this.selectQueuesCount,
					params = { qUE, i -> i[qUE.P_QueueSetId] = row[rowUE.C_QueueSet] },
					result = { qUE, o -> return@selectOneOrError o[qUE.C_HasQueries] }
				)

				return hasRows == (row[rowUE.C_Queue] == null)
			}
		}
	}

	private class _Constructor(
		private val usersTable: Table<T_Users, *>,
		private val queueSetsTable: Table<T_QueueSets, *>,
		private val queuesTable: Table<T_Queues, *>
	) : TableConstructor<T_User2Queue> {
		override fun createTable(context: TableConstructor.Scope<T_User2Queue>) =
			T_User2Queue(context, this.usersTable, this.queueSetsTable, this.queuesTable)

	}

	companion object Constructor {
		operator fun invoke(usersTable: Table<T_Users, *>, queueSetsTable: Table<T_QueueSets, *>, queuesTable: Table<T_Queues, *>): TableConstructor<T_User2Queue> =
			_Constructor(usersTable, queueSetsTable, queuesTable)
	}
}