package ru.landgrafhomyak.bgtu.db0.storage.serdha

import ru.landgrafhomyak.bgtu.db0.storage.api.ModeratorPermissions
import ru.landgrafhomyak.bgtu.db0.storage.api.Storage
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.D_RemoveUserFromQueue
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.I_AddQueueSet
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.I_AddQueues
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.I_AddUserToQueue
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.S_FindFreeOrdinalInQueue
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.S_FindNeighbourQueue
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.S_FindQueueSet
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.S_FindUserInQueue
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.S_FirstQueueInSet
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.U_ChangeQueueState
import ru.landgrafhomyak.bgtu.db0.storage.serdha.queries.U_MoveUserInQueue
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.executeNoReturn
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.insertMany
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.selectOneOrError
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.selectOneOrNull
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.transaction
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.M_V0
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.T_QueueSets
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.T_Queues
import ru.landgrafhomyak.db.serdha0.api.module.Module
import ru.landgrafhomyak.db.serdha0.api.queries.SelectQuery
import ru.landgrafhomyak.db.serdha0.api.runtime.Executor
import ru.landgrafhomyak.db.serdha0.api.runtime.Transaction
import ru.landgrafhomyak.db.serdha0.api.table.RowId
import ru.landgrafhomyak.bgtu.db0.storage.api.Storage as IStorage

@Suppress("PrivatePropertyName")
class SerdhaStorage(module: Module<M_V0>, private val executor: Executor) : IStorage {

	private val _q_addQueueSet = module.createQuery(I_AddQueueSet(module))
	private val _q_addQueues = module.createQuery(I_AddQueues(module))

	/*override*/ suspend fun allocQueue(chat: Long, message: Long, setName: String, queueNames: List<String>) = this.executor.transaction { t ->
		val qsRowId = t.selectOneOrError(
			compiledQuery = this._q_addQueueSet,
			params = { qUE, p ->
				p[qUE.P_TgCid] = chat
				p[qUE.P_TgMid] = message.toULong()
				p[qUE.P_Name] = setName
			},
			result = { qUE, r ->
				return@selectOneOrError r[qUE.C_RowId]
			}
		)

		var i: UByte = 0u
		t.insertMany(this._q_addQueues, queueNames) { qUE, p, e ->
			p[qUE.P_QueueSet] = qsRowId
			p[qUE.P_Ordinal] = i++
			p[qUE.P_Name] = e
		}
	}

	private val _q_changeQueueState = module.createQuery(U_ChangeQueueState(module))

	@Suppress("FunctionName")
	private suspend fun _closeOrOpenQueue(chat: Long, message: Long, newIsOpen: Boolean): Storage.Result_QueueCloseOrOpen = this.executor.transaction { t ->
		val qs = t.selectOneOrNull(
			compiledQuery = this._q_changeQueueState,
			params = { qUE, p ->
				p[qUE.P_TgCid] = chat
				p[qUE.P_TgMid] = message.toULong()
				p[qUE.P_NewState] = newIsOpen
			},
			// result = FoundQueue::fromRow
			result = { qUE, r ->
				if (r[qUE.C_OldIsOpen] == newIsOpen)
					return@transaction Storage.Result_QueueCloseOrOpen.ALREADY_IN_STATE
				return@selectOneOrNull FoundQueue.fromRow(qUE, r)
			}
		)

		if (qs == null)
			return@transaction Storage.Result_QueueCloseOrOpen.QUEUE_NOT_FOUND

		return@transaction Storage.Result_QueueCloseOrOpen.SUCCESSFUL(this._formatQueueSet(qs))
	}

	/*override*/ suspend fun closeQueue(chat: Long, message: Long): Storage.Result_QueueCloseOrOpen =
		this._closeOrOpenQueue(chat, message, false)

	/*override*/ suspend fun reopenQueue(chat: Long, message: Long): Storage.Result_QueueCloseOrOpen =
		this._closeOrOpenQueue(chat, message, true)


	private val _q_findQueueSet = module.createQuery(S_FindQueueSet(module))

	@Suppress("FunctionName")
	private suspend fun _findQueueSet(transaction: Transaction, chat: Long, message: Long): FoundQueue? =
		transaction.selectOneOrNull(
			compiledQuery = this._q_findQueueSet.byMessage,
			params = { ue, r ->
				r[ue.P_TgCid] = chat
				r[ue.P_TgMid] = message.toULong()
			},
			result = FoundQueue::fromRow
		)

	private val _q_findUserInQueue = module.createQuery(S_FindUserInQueue(module))

	@Suppress("FunctionName")
	private suspend fun _findUserInQueue(transaction: Transaction, qs: FoundQueue, user: Long, displayName: String): FoundUserInQueue =
		transaction.selectOneOrError(
			compiledQuery = this._q_findUserInQueue,
			params = { ue, r ->
				r[ue.P_QueueSet] = qs.rowId
				r[ue.P_TgUid] = user.toULong()
				r[ue.P_UDisplayName] = displayName
			},
			result = FoundUserInQueue::fromRow
		)

	private val _q_findFirstQueue = module.createQuery(S_FirstQueueInSet(module))
	private val _q_findOrdinal = module.createQuery(S_FindFreeOrdinalInQueue(module))
	private val _q_addUser2Queue = module.createQuery(I_AddUserToQueue(module))

	/*override*/ suspend fun addUserToQueue(chat: Long, message: Long, user: Long, userDisplayName: String): Storage.Result_AddOrRemoveUser = this.executor.transaction { transaction ->
		val qs = this._findQueueSet(transaction, chat, message) ?: return@transaction Storage.Result_AddOrRemoveUser.QUEUE_NOT_FOUND

		if (!qs.isOpen) return@transaction Storage.Result_AddOrRemoveUser.QUEUE_IS_FINAL

		val u = this._findUserInQueue(transaction, qs, user, userDisplayName)
		if (u.ordinal != null) return@transaction Storage.Result_AddOrRemoveUser.ALREADY_IN_STATE

		val q = transaction.selectOneOrNull(
			compiledQuery = this._q_findFirstQueue,
			params = { ue, r -> r[ue.P_QueueSet] = qs.rowId },
			result = { ue, r -> return@selectOneOrNull r[ue.C_Queue] }
		)

		val o = this._q_findOrdinal.findFirstOrdinal(transaction, 0u, qs.rowId, q) ?: return@transaction Storage.Result_AddOrRemoveUser.QUEUE_OVERFLOW

		transaction.executeNoReturn(this._q_addUser2Queue) { x, r ->
			r[x.P_QueueSet] = qs.rowId
			r[x.P_Queue] = q
			r[x.P_User] = u.userRowId
			r[x.P_Ordinal] = o
		}

		return@transaction Storage.Result_AddOrRemoveUser.SUCCESSFUL(this._formatQueueSet(qs))
	}

	private val _q_removeUserFromQueue = module.createQuery(D_RemoveUserFromQueue(module))

	/*override*/ suspend fun removeUserFromQueue(chat: Long, message: Long, user: Long, userDisplayName: String): Storage.Result_AddOrRemoveUser = this.executor.transaction { transaction ->
		val qs = this._findQueueSet(transaction, chat, message) ?: return@transaction Storage.Result_AddOrRemoveUser.QUEUE_NOT_FOUND

		if (!qs.isOpen) return@transaction Storage.Result_AddOrRemoveUser.QUEUE_IS_FINAL

		val u = this._findUserInQueue(transaction, qs, user, userDisplayName)
		if (u.ordinal == null) return@transaction Storage.Result_AddOrRemoveUser.ALREADY_IN_STATE

		transaction.executeNoReturn(this._q_removeUserFromQueue) { x, r ->
			r[x.P_RowId] = u.linkRowId
		}

		return@transaction Storage.Result_AddOrRemoveUser.SUCCESSFUL(this._formatQueueSet(qs))
	}

	private val _q_findNeighbourQueue = module.createQuery(S_FindNeighbourQueue(module))
	private val _q_moveUserInQueue = module.createQuery(U_MoveUserInQueue(module))

	@Suppress("FunctionName")
	private suspend fun _moveUserToNeighbourQueue(chat: Long, message: Long, user: Long, userDisplayName: String, query: SelectQuery.Simple<S_FindNeighbourQueue.Query>): Storage.Result_MoveToNextOrPrevQueue = this.executor.transaction { transaction ->
		val qs = this._findQueueSet(transaction, chat, message) ?: return@transaction Storage.Result_MoveToNextOrPrevQueue.QUEUE_NOT_FOUND

		if (!qs.isOpen) return@transaction Storage.Result_MoveToNextOrPrevQueue.QUEUE_IS_FINAL

		val u = this._findUserInQueue(transaction, qs, user, userDisplayName)

		if (u.ordinal == null)
			return@transaction Storage.Result_MoveToNextOrPrevQueue.NOT_IN_QUEUE
		if (u.queueRef == null)
			return@transaction Storage.Result_MoveToNextOrPrevQueue.ON_EDGE

		val qo = transaction.selectOneOrError(
			this._q_findNeighbourQueue.getOrdinal,
			params = { x, r -> r[x.P_RowId] = u.queueRef },
			result = { x, r -> return@selectOneOrError r[x.C_Ordinal] }
		)

		val q = transaction.selectOneOrNull(
			compiledQuery = query,
			params = { ue, r -> r[ue.P_QueueSet] = qs.rowId; r[ue.P_OldOrdinal] = qo },
			result = { ue, r -> return@selectOneOrNull r[ue.C_NewQueue] }
		)

		if (q == null)
			return@transaction Storage.Result_MoveToNextOrPrevQueue.ON_EDGE

		val o = this._q_findOrdinal.findFirstOrdinal(transaction, 0u, qs.rowId, q) ?: return@transaction Storage.Result_MoveToNextOrPrevQueue.QUEUE_OVERFLOW

		transaction.executeNoReturn(this._q_moveUserInQueue) { x, r ->
			r[x.P_RowId] = u.linkRowId
			r[x.P_NewQueue] = q
			r[x.P_NewOrdinal] = o
		}

		return@transaction Storage.Result_MoveToNextOrPrevQueue.SUCCESSFUL(this._formatQueueSet(qs))
	}

	/*override*/ suspend fun moveUserToNextQueue(chat: Long, message: Long, user: Long, userDisplayName: String): Storage.Result_MoveToNextOrPrevQueue =
		this._moveUserToNeighbourQueue(chat, message, user, userDisplayName, this._q_findNeighbourQueue.next)

	/*override*/ suspend fun moveUserToPrevQueue(chat: Long, message: Long, user: Long, userDisplayName: String): Storage.Result_MoveToNextOrPrevQueue =
		this._moveUserToNeighbourQueue(chat, message, user, userDisplayName, this._q_findNeighbourQueue.prev)

	@Suppress("FunctionName")
	private suspend fun _moveUserInQueue(
		chat: Long, message: Long, user: Long, userDisplayName: String,
		query: suspend (t: Transaction, o: UByte, qs: RowId<T_QueueSets>, q: RowId<T_Queues>?) -> UByte?
	): Storage.Result_MoveToStartOrEnd = this.executor.transaction { transaction ->
		val qs = this._findQueueSet(transaction, chat, message) ?: return@transaction Storage.Result_MoveToStartOrEnd.QUEUE_NOT_FOUND

		if (!qs.isOpen) return@transaction Storage.Result_MoveToStartOrEnd.QUEUE_IS_FINAL

		val u = this._findUserInQueue(transaction, qs, user, userDisplayName)

		val o = query(transaction, u.ordinal ?: return@transaction Storage.Result_MoveToStartOrEnd.NOT_IN_QUEUE, qs.rowId, u.queueRef) ?: return@transaction Storage.Result_MoveToStartOrEnd.ON_EDGE

		transaction.executeNoReturn(this._q_moveUserInQueue) { x, r ->
			r[x.P_RowId] = u.linkRowId
			r[x.P_NewQueue] = u.queueRef
			r[x.P_NewOrdinal] = o
		}

		return@transaction Storage.Result_MoveToStartOrEnd.SUCCESSFUL(this._formatQueueSet(qs))
	}

	/*override*/ suspend fun moveUserToStartOfQueue(chat: Long, message: Long, user: Long, displayName: String) =
		this._moveUserInQueue(chat, message, user, displayName, this._q_findOrdinal::findOrdinalBefore)

	/*override*/ suspend fun moveUserToEndOfQueue(chat: Long, message: Long, user: Long, displayName: String) =
		this._moveUserInQueue(chat, message, user, displayName, this._q_findOrdinal::findOrdinalAfter)


	override fun addModerator(chat: Long, user: Long): Boolean {
		TODO("Not yet implemented")
	}

	override fun removeModerator(chat: Long, message: Long): Boolean {
		TODO("Not yet implemented")
	}

	override fun getPermissions(chat: Long, user: Long): ModeratorPermissions? {
		TODO("Not yet implemented")
	}

	override fun setPermission(chat: Long, user: Long, permissionType: ModeratorPermissions.PermissionType, value: Boolean): Storage.Result_SetModeratorPermission {
		TODO("Not yet implemented")
	}
}