package ru.landgrafhomyak.bgtu.db0.storage.serdha.queries

import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.U8
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp.selectOneOrNull
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.M_V0
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.T_QueueSets
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.T_Queues
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.T_User2Queue
import ru.landgrafhomyak.db.serdha0.api.misc.Column
import ru.landgrafhomyak.db.serdha0.api.misc.DatabaseType
import ru.landgrafhomyak.db.serdha0.api.misc.FiltersScope
import ru.landgrafhomyak.db.serdha0.api.misc.IntermediateColumn
import ru.landgrafhomyak.db.serdha0.api.misc.QueryParam
import ru.landgrafhomyak.db.serdha0.api.misc.Reducer
import ru.landgrafhomyak.db.serdha0.api.misc._Selectable
import ru.landgrafhomyak.db.serdha0.api.module.Module
import ru.landgrafhomyak.db.serdha0.api.queries.SelectQuery
import ru.landgrafhomyak.db.serdha0.api.queries._Query
import ru.landgrafhomyak.db.serdha0.api.runtime.InputRow
import ru.landgrafhomyak.db.serdha0.api.runtime.OutputRow
import ru.landgrafhomyak.db.serdha0.api.runtime.Transaction
import ru.landgrafhomyak.db.serdha0.api.table.RowId

@Suppress("ClassName", "PropertyName", "MemberVisibilityCanBePrivate", "PrivatePropertyName")
class S_FindFreeOrdinalInQueue(private val module: Module<M_V0>) : _Query.Constructor<S_FindFreeOrdinalInQueue.EntryPoints> {
	override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope) =
		EntryPoints(this.module, outQueries)


	@Suppress("PrivatePropertyName", "FoldInitializerAndIfToElvis")
	class EntryPoints(module: Module<M_V0>, outQueries: _Query.Constructor.Scope) {
		private val toFront_Q = outQueries.selectReducing(Query.Constructor(module.uExt.user2queue, Dec) { t -> t.C_Queue })
		private val toFront_QS = outQueries.selectReducing(Query.Constructor(module.uExt.user2queue, Dec) { t -> t.C_QueueSet })
		private val toBack_Q = outQueries.selectReducing(Query.Constructor(module.uExt.user2queue, Inc) { t -> t.C_Queue })
		private val toBack_QS = outQueries.selectReducing(Query.Constructor(module.uExt.user2queue, Inc) { t -> t.C_QueueSet })
		private val first_Q = outQueries.selectReducing(Query.Constructor(module.uExt.user2queue, Inc.Inclusive) { t -> t.C_Queue })
		private val first_QS = outQueries.selectReducing(Query.Constructor(module.uExt.user2queue, Inc.Inclusive) { t -> t.C_QueueSet })

		@Suppress("DuplicatedCode", "FunctionName")
		private suspend fun _execute(
			t: Transaction,
			qq: SelectQuery.Reducing<Query<RowId<T_Queues>?, DatabaseType.Nullable<RowId<T_Queues>, DatabaseType.ROW_ID<T_Queues>>>>,
			qqs: SelectQuery.Reducing<Query<RowId<T_QueueSets>, DatabaseType.ROW_ID<T_QueueSets>>>,
			o: UByte,
			qs: RowId<T_QueueSets>,
			q: RowId<T_Queues>?
		): UByte? {
			if (q == null) {
				return t.selectOneOrNull(
					qqs,
					params = { x, r -> r[x.P_Ordinal] = o;r[x._P_Selector] = qs },
					result = { x, r -> return@selectOneOrNull r[x.C_Ordinal] }
				)
			} else {
				return t.selectOneOrNull(
					qq,
					params = { x, r -> r[x.P_Ordinal] = o;r[x._P_Selector] = q },
					result = { x, r -> return@selectOneOrNull r[x.C_Ordinal] }
				)
			}
		}

		suspend fun findOrdinalBefore(t: Transaction, o: UByte, qs: RowId<T_QueueSets>, q: RowId<T_Queues>?): UByte? {
			val f = this._execute(t, this.toFront_Q, this.toFront_QS, o, qs, q)
			if (f == null) return if (o.toUInt() == 0u) null else (o - 1u).toUByte()
			if (f.toUInt() == 0u) return null
			return (f - 1u).toUByte()
		}

		suspend fun findOrdinalAfter(t: Transaction, o: UByte, qs: RowId<T_QueueSets>, q: RowId<T_Queues>?): UByte? {
			val f = this._execute(t, this.toBack_Q, this.toBack_QS, o, qs, q)
			if (f == null) return if (o.toUInt() == 255u) null else (o + 1u).toUByte()
			if (f.toUInt() == 255u) return null
			return (f + 1u).toUByte()
		}

		suspend fun findFirstOrdinal(t: Transaction, o: UByte, qs: RowId<T_QueueSets>, q: RowId<T_Queues>?): UByte? {
			val f = this._execute(t, this.first_Q, this.first_QS, o, qs, q)
			if (f == null) return 0u
			if (f.toUInt() == 255u) return null
			return (f + 1u).toUByte()
		}
	}


	private abstract class _OrdinalSelectorReducerConstructor(
		val ordinalCompOp: FiltersScope.ComparisonType,
		val order: SelectQuery.Order
	) {
		abstract fun <sRT, sDT : DatabaseType<sRT>> setup(
			c: SelectQuery.Reducing.Constructor.Scope<Query<sRT, sDT>, T_User2Queue>,
			o: IntermediateColumn<UByte, U8, Query<sRT, sDT>>,
			s: QueryParam<UByte, U8, Query<sRT, sDT>>
		)
	}

	private class Inc private constructor(context: Reducer.Constructor.Scope<Inc>) : Reducer.Action<Inc> {
		private val C_Ordinal = context.bidirectionalColumn("last_ordinal", context.types.U8)

		override fun calculate(acc: OutputRow<Inc>, row: OutputRow<Inc>, newAcc: InputRow<Inc>) {
			val new = row[this.C_Ordinal]
			if (acc[this.C_Ordinal].toUInt() + 1u == new.toUInt())
				newAcc[this.C_Ordinal] = new
		}

		companion object : _OrdinalSelectorReducerConstructor(FiltersScope.ComparisonType.GREATER, SelectQuery.Order.ASC), Reducer.Constructor<Inc> {
			override fun createReducer(context: Reducer.Constructor.Scope<Inc>) = Inc(context)
			override fun <sRT, sDT : DatabaseType<sRT>> setup(
				c: SelectQuery.Reducing.Constructor.Scope<Query<sRT, sDT>, T_User2Queue>,
				o: IntermediateColumn<UByte, U8, Query<sRT, sDT>>,
				s: QueryParam<UByte, U8, Query<sRT, sDT>>
			) {
				c.reduceRows(c.createReducer(this)) { l -> l.linkAndInitialize(l.uExt.C_Ordinal, o, s) }
			}
		}

		object Inclusive : _OrdinalSelectorReducerConstructor(FiltersScope.ComparisonType.GREATER_OR_EQUALS, SelectQuery.Order.ASC) {
			override fun <sRT, sDT : DatabaseType<sRT>> setup(
				c: SelectQuery.Reducing.Constructor.Scope<Query<sRT, sDT>, T_User2Queue>,
				o: IntermediateColumn<UByte, U8, Query<sRT, sDT>>,
				s: QueryParam<UByte, U8, Query<sRT, sDT>>
			) {
				c.reduceRows(c.createReducer(Inc)) { l -> l.linkAndInitialize(l.uExt.C_Ordinal, o, 0u) }
			}
		}
	}


	private class Dec private constructor(context: Reducer.Constructor.Scope<Dec>) : Reducer.Action<Dec> {
		private val C_Ordinal = context.bidirectionalColumn("last_ordinal", context.types.U8)

		override fun calculate(acc: OutputRow<Dec>, row: OutputRow<Dec>, newAcc: InputRow<Dec>) {
			val new = row[this.C_Ordinal]
			if (acc[this.C_Ordinal].toUInt() - 1u == new.toUInt())
				newAcc[this.C_Ordinal] = new
		}

		companion object : _OrdinalSelectorReducerConstructor(FiltersScope.ComparisonType.LESS, SelectQuery.Order.DESC), Reducer.Constructor<Dec> {
			override fun createReducer(context: Reducer.Constructor.Scope<Dec>) = Dec(context)

			override fun <sRT, sDT : DatabaseType<sRT>> setup(
				c: SelectQuery.Reducing.Constructor.Scope<Query<sRT, sDT>, T_User2Queue>,
				o: IntermediateColumn<UByte, U8, Query<sRT, sDT>>,
				s: QueryParam<UByte, U8, Query<sRT, sDT>>
			) {
				c.reduceRows(c.createReducer(this)) { l -> l.linkAndInitialize(l.uExt.C_Ordinal, o, s) }
			}
		}
	}

	private class Query<sRT, sDT : DatabaseType<sRT>> private constructor(
		context: SelectQuery.Reducing.Constructor.Scope<Query<sRT, sDT>, T_User2Queue>,
		reducer: _OrdinalSelectorReducerConstructor,
		selectorColumn: (T_User2Queue) -> Column<sRT, sDT, T_User2Queue>
	) {
		val P_Ordinal = context.queryParam("ordinal", context.src.uExt.C_Ordinal.type)
		val _P_Selector = context.queryParam("_selector", selectorColumn(context.src.uExt).type)
		val C_Ordinal: Column<UByte, U8, Query<sRT, sDT>>

		init {
			val ordinalColumn = context.selectColumnFromSubquery(context.src.uExt.C_Ordinal)
			this.C_Ordinal = context.returnColumn("ordinal", ordinalColumn)
			reducer.setup(context, ordinalColumn, this.P_Ordinal)
			context.filters.compareWithParam(ordinalColumn, reducer.ordinalCompOp, this.P_Ordinal)
			context.orderBy(reducer.order, ordinalColumn)
			context.filters.compareWithParam(context.selectColumnFromSubquery(selectorColumn(context.src.uExt)), FiltersScope.ComparisonType.EQUALS, this._P_Selector)
		}

		class Constructor<sRT, sDT : DatabaseType<sRT>>(
			override val source: _Selectable<T_User2Queue>,
			private val reducer: _OrdinalSelectorReducerConstructor,
			private val selectorColumn: (T_User2Queue) -> Column<sRT, sDT, T_User2Queue>
		) : SelectQuery.Reducing.Constructor<Query<sRT, sDT>, T_User2Queue> {
			override fun createSelectWithReduce(context: SelectQuery.Reducing.Constructor.Scope<Query<sRT, sDT>, T_User2Queue>) =
				Query(context, this.reducer, this.selectorColumn)
		}
	}
}