package ru.landgrafhomyak.bgtu.db0.storage.serdha.queries

import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.M_V0
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.T_Queues
import ru.landgrafhomyak.db.serdha0.api.misc.FiltersScope
import ru.landgrafhomyak.db.serdha0.api.misc._Selectable
import ru.landgrafhomyak.db.serdha0.api.module.Module
import ru.landgrafhomyak.db.serdha0.api.queries.SelectQuery
import ru.landgrafhomyak.db.serdha0.api.queries._Query

@Suppress("ClassName", "PropertyName", "MemberVisibilityCanBePrivate")
internal class S_FindNeighbourQueue(private val module: Module<M_V0>) : _Query.Constructor<S_FindNeighbourQueue.Queries> {
	override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope) =
		Queries(
			getOrdinal = outQueries.select(S_GetQueueOrdinal.Constructor(this.module.uExt.queues)),
			prev = outQueries.select(Query.Constructor(this.module.uExt.queues, SelectQuery.Order.DESC, FiltersScope.ComparisonType.LESS)),
			next = outQueries.select(Query.Constructor(this.module.uExt.queues, SelectQuery.Order.ASC, FiltersScope.ComparisonType.GREATER))
		)

	class Queries(
		val getOrdinal: SelectQuery.Simple<S_GetQueueOrdinal>,
		val prev: SelectQuery.Simple<Query>,
		val next: SelectQuery.Simple<Query>
	)


	class S_GetQueueOrdinal(
		context: SelectQuery.Simple.Constructor.Scope<S_GetQueueOrdinal, T_Queues>
	) {
		val P_RowId = context.queryParam("row_id", context.src.uExt.C_RowId.type)
		val C_Ordinal = context.returnColumn("ordinal", context.selectColumnFromSubquery(context.src.uExt.C_Ordinal))

		class Constructor(override val source: _Selectable<T_Queues>) : SelectQuery.Simple.Constructor<S_GetQueueOrdinal, T_Queues> {
			override fun createSelect(context: SelectQuery.Simple.Constructor.Scope<S_GetQueueOrdinal, T_Queues>) = S_GetQueueOrdinal(context)
		}
	}

	class Query private constructor(
		context: SelectQuery.Simple.Constructor.Scope<Query, T_Queues>,
		order: SelectQuery.Order,
		operator: FiltersScope.ComparisonType
	) {
		val P_QueueSet = context.queryParam("queue_set", context.src.uExt.C_QueueSet.type)
		val P_OldOrdinal = context.queryParam("old_ordinal", context.src.uExt.C_Ordinal.type)
		val C_NewQueue = context.returnColumn("new_queue", context.selectColumnFromSubquery(context.src.uExt.C_RowId))

		init {
			val ordinalColumn = context.selectColumnFromSubquery(context.src.uExt.C_Ordinal)
			context.filters.compareWithParam(
				column = ordinalColumn,
				param = this.P_OldOrdinal,
				operator = operator
			)
			context.orderBy(order, ordinalColumn)
		}

		class Constructor(
			override val source: _Selectable<T_Queues>,
			private val order: SelectQuery.Order,
			private val operator: FiltersScope.ComparisonType
		) : SelectQuery.Simple.Constructor<Query, T_Queues> {
			override fun createSelect(context: SelectQuery.Simple.Constructor.Scope<Query, T_Queues>) =
				Query(context, this.order, this.operator)
		}
	}
}