package ru.landgrafhomyak.bgtu.db0.storage.serdha.queries

import ru.landgrafhomyak.bgtu.db0.storage.serdha.FoundQueue
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.M_V0
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.T_Chats
import ru.landgrafhomyak.bgtu.db0.storage.serdha.schema.v0.T_QueueSets
import ru.landgrafhomyak.db.serdha0.api.misc.Column
import ru.landgrafhomyak.db.serdha0.api.misc.FiltersScope
import ru.landgrafhomyak.db.serdha0.api.misc._Selectable
import ru.landgrafhomyak.db.serdha0.api.module.Module
import ru.landgrafhomyak.db.serdha0.api.queries.SelectQuery
import ru.landgrafhomyak.db.serdha0.api.queries._Query

@Suppress("ClassName", "PropertyName", "MemberVisibilityCanBePrivate")
internal class S_FindQueueSet(private val module: Module<M_V0>) : _Query.Constructor<S_FindQueueSet.Queries> {
	override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope) =
		Queries(
			byMessage = outQueries.selectWithJoin(
				S_FindQueueSetByMessage.Constructor(
					this.module.uExt.queueSets,
					internalQueries.select(S_FindChat.Constructor(this.module.uExt.chats))
				)
			),
			byId = outQueries.select(S_FindQueueSetById.Constructor(this.module.uExt.queueSets))
		)

	class Queries(
		val byMessage: SelectQuery.WithJoin<S_FindQueueSetByMessage>,
		val byId: SelectQuery.Simple<S_FindQueueSetById>
	)

	class S_FindChat private constructor(context: SelectQuery.Simple.Constructor.Scope<S_FindChat, T_Chats>) {
		val P_TgCid = context.queryParam("tg_cid", context.src.uExt.C_TgCid.type)
		val C_RowId = context.returnColumn("row_id", context.selectColumnFromSubquery(context.src.uExt.C_RowId))

		init {
			context.filters.compareWithParam(
				column = context.selectColumnFromSubquery(context.src.uExt.C_TgCid),
				param = this.P_TgCid,
				operator = FiltersScope.ComparisonType.EQUALS
			)
		}

		class Constructor(override val source: _Selectable<T_Chats>) : SelectQuery.Simple.Constructor<S_FindChat, T_Chats> {
			override fun createSelect(context: SelectQuery.Simple.Constructor.Scope<S_FindChat, T_Chats>) =
				S_FindChat(context)
		}
	}

	class S_FindQueueSetByMessage private constructor(context: SelectQuery.WithJoin.Constructor.Scope<S_FindQueueSetByMessage, T_QueueSets, S_FindChat>) : FoundQueue.QueueSetRow<S_FindQueueSetByMessage> {
		val P_TgCid = context.queryParam("tg_cid", context.rsq.uExt.P_TgCid.type)
		val P_TgMid = context.queryParam("tg_mid", context.lsq.uExt.C_TgMid.type)
		override val C_QsRowId = context.returnColumn("row_id", context.selectColumnFromLeft(context.lsq.uExt.C_RowId))
		override val C_QsName = context.returnColumn("name", context.selectColumnFromLeft(context.lsq.uExt.C_Name))
		override val C_QsIsOpen = context.returnColumn("is_open", context.selectColumnFromLeft(context.lsq.uExt.C_IsOpen))

		init {
			context.rightSubqueryParams.redirect(context.rsq.uExt.P_TgCid, this.P_TgCid)
		}

		class Constructor(override val left: _Selectable<T_QueueSets>, override val right: _Selectable<S_FindChat>) : SelectQuery.WithJoin.Constructor<S_FindQueueSetByMessage, T_QueueSets, S_FindChat> {
			override val joinType: SelectQuery.JoinType = SelectQuery.JoinType(
				SelectQuery.JoinType.LeftNull.NULL_NOT_EXPECTED,
				SelectQuery.JoinType.Join.LEFT_TO_FIRST_RIGHT
			)
			override val referenceColumn: Column<*, *, T_QueueSets> = this.left.uExt.C_Chat
			override val targetColumn: Column<*, *, S_FindChat> = this.right.uExt.C_RowId

			override fun createSelectWithJoin(context: SelectQuery.WithJoin.Constructor.Scope<S_FindQueueSetByMessage, T_QueueSets, S_FindChat>) =
				S_FindQueueSetByMessage(context)
		}
	}


	class S_FindQueueSetById private constructor(context: SelectQuery.Simple.Constructor.Scope<S_FindQueueSetById, T_QueueSets>) : FoundQueue.QueueSetRow<S_FindQueueSetById> {
		val P_RowId = context.queryParam("tg_cid", context.src.uExt.C_RowId.type)
		override val C_QsRowId = context.returnColumn("row_id", context.selectColumnFromSubquery(context.src.uExt.C_RowId))
		override val C_QsName = context.returnColumn("name", context.selectColumnFromSubquery(context.src.uExt.C_Name))
		override val C_QsIsOpen = context.returnColumn("is_open", context.selectColumnFromSubquery(context.src.uExt.C_IsOpen))

		init {
			context.filters.compareWithParam(
				column = context.selectColumnFromSubquery(context.src.uExt.C_RowId),
				param = this.P_RowId,
				operator = FiltersScope.ComparisonType.EQUALS
			)
		}

		class Constructor(override val source: _Selectable<T_QueueSets>) : SelectQuery.Simple.Constructor<S_FindQueueSetById, T_QueueSets> {
			override fun createSelect(context: SelectQuery.Simple.Constructor.Scope<S_FindQueueSetById, T_QueueSets>) =
				S_FindQueueSetById(context)
		}
	}
}